/*
	Copyright (C) 2020-2022 Lewis ANESA.

	This file is part of ladspa-tool-kit

	ladspa-tool-kit is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	ladspa-tool-kit is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ladspa-tool-kit.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <execinfo.h>
#include <sys/epoll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#define UNW_LOCAL_ONLY
#include <libunwind.h>

#include "../include/instance.h"
#include "../include/utils.h"

uint32_t LTKRand(uint32_t seed) {
	return seed * 1103515245 + 12345;
}

uint64_t LTKHash(uint64_t seed) {
	/* Inversion variables */
/*	uint64_t inverse = 0; */
/*	int x = 64; */
	uint64_t inverse;
	int x = 8;
	uint8_t *in = (uint8_t *) & seed;
	uint8_t *out = (uint8_t *) & inverse;
	uint8_t lookup[] = {
		0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
		0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
		0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8,
		0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
		0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
		0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
		0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC,
		0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
		0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2,
		0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
		0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA,
		0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
		0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6,
		0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
		0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
		0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
		0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1,
		0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
		0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9,
		0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
		0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5,
		0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
		0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED,
		0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
		0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
		0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
		0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB,
		0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
		0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7,
		0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
		0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF,
		0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
	};

	/* Cellular automaton variables */
	uint64_t shifted[3];
	uint64_t result[3];

	/* Inversion part */
/*	while(x--) if(hash_in & (1LU << x)) inverse |= 1LU << (63 - x); */
	while(x--)
		out[7 - x] = lookup[in[x]];

	/* Cellular automaton part (rule 30 shifted by 21 on left) */
	x = 3;

	while(x--)
		shifted[x] = (seed << (20 + x)) | (seed >> (44 - x));
	result[1] = ~shifted[2] & (shifted[0] | shifted[1]);
	result[2] = ~shifted[0] & ~shifted[1] & shifted[2];
	result[0] = result[1] | result[2];

	return (seed * seed) ^ (inverse ^ result[0]);
}

void *LTKArray(size_t cell, ...) {
	uint64_t dim[256];
	va_list arg;
	uint8_t *ret;
	uint8_t *dst;
	void **src;
	size_t it = 0;
	size_t size = 0;
	uint64_t acc = 1;

	va_start(arg, cell);
	while ((dim[it++] = va_arg(arg, unsigned int))) ;
	va_end(arg);

	if (dim[0] && cell) {
		for (it = 0; dim[it]; it++) {
			acc *= dim[it];
			size += acc;
		}

		size = (size * sizeof(void *)) + (acc * cell);
		acc = 1;
		if (!(ret = calloc(1, size))) {
			LTK_LOG_ERROR("Cannot allocte array : %s",
				      strerror(errno));

			return 0;
		}
		src = (void **)ret;
		dst = ret + (dim[0] * sizeof(void *));

		for (it = 0; dim[it]; it++) {
			acc *= dim[it];

			for (size = 0; size < acc; size++) {
				*(src++) = (void *) dst;

				if (dim[it + 1])
					dst += (dim[it + 1] * sizeof(void *));
				else
					dst += cell;
			}
		}
	} else
		ret = 0;

	return (void *)ret;
}

void *LTKmemcpy(uint8_t *dest, uint8_t *src, size_t size) {
	size_t gap = dest - src;
	if(size <= 0) return dest;
	if(gap < 0) gap = -gap;

	if(gap > size) memmove(dest, src, size);
	else memcpy(dest, src, size);

	return dest;
}

char *LTKBacktrace(char *before_str, char *indent_str
		, char *end_str, char *after_str) {
	static char buffer[BUFSIZ] = {0};
	char *ptr = buffer;
	int x = 0;

	unw_context_t context;
	unw_cursor_t cursor;
	unw_word_t offset;

	unw_getcontext(&context);
	unw_init_local(&cursor, &context);

	if(before_str)
		ptr += sprintf(ptr, "%s", before_str);

	while(unw_step(&cursor) > 0) {
		int indent = x;

		while(indent--) {
			if(indent) {
				if (indent_str)
					ptr += sprintf(ptr, "%s", indent_str);
			}
			else {
				if (end_str)
					ptr += sprintf(ptr, "%s", end_str);
				else if (indent_str)
					ptr += sprintf(ptr, "%s", indent_str);
			}
		}

		unw_get_proc_name(&cursor, ptr, 255, &offset);
		ptr[255] = 0;

		while(ptr[0])
			ptr++;

		ptr[0] = '\n';
		ptr++;

		x++;
	}

	if(after_str)
		ptr += sprintf(ptr, "%s", after_str);
	ptr[0] = 0;

	return buffer;
}
